/******************************************
 * @author lkl <klk0@qq.com>
 * @copyright zk 2019/8/2
 * @doc 关联组件.
 * @end
 ******************************************/
const Type = cc.Enum({
    NONE: 0,
    WIDTH_WIDTH: 1,//宽对宽
    HEIGHT_HEIGHT: 2,//高对高
    LEFT_LEFT: 3,//左对左
    LEFT_CENTER: 4,//左对中
    LEFT_RIGHT: 5,//左对右
    RIGHT_RIGHT: 6,//右对右
    RIGHT_CENTER: 7,//右对中
    RIGHT_LEFT: 8,//右对左
    TOP_TOP: 9,//顶对顶
    TOP_CENTER: 10,//顶对中
    TOP_BOTTOM: 11,//顶对底
    BOTTOM_BOTTOM: 12,//底对底
    BOTTOM_CENTER: 13,//底对中
    BOTTOM_TOP: 14,//底对顶
    HORIZONTAL_CENTER: 15,//左右居中
    VERTICAL_CENTER: 16,//上下居中
});

cc.Class({
    editor: {
        menu: "自定义组件/Relation",
        executeInEditMode: true,
        executionOrder: -5002,
    },
    extends: cc.Component,

    properties: {
        target: {       //关联目标
            type: cc.Node,
            default: null,
            notify: function () {
                if (this.target) {
                    this._unregisterEvent();
                    this._registerEvent();
                }
                this._initTargetInfo();
            },
        },
        relations: {    //关联列表
            default: [],
            type: Type,
            // notify: function () {
            // }
        },
        // percentageUsed: false,  //使用百分比（暂时还没做。。）
    },

    onLoad() {
        this._initNodeInfo();
        this._initTargetInfo();
        this._registerEvent();
    },

    onEnable() {
        this._initTargetInfo();
        this._initNodeInfo();
    },

    onDisable() {
        this._unregisterEvent();
    },

    _registerEvent() {
        let t = this;
        cc.director.on(cc.Director.EVENT_AFTER_UPDATE, t._onEvAfterUpdate, t);
        t.node.on(cc.Node.EventType.POSITION_CHANGED, t._initNodeInfo, t);
        t.node.on(cc.Node.EventType.SIZE_CHANGED, t._initNodeInfo, t);
        if (t.target) {
            t.target.on(cc.Node.EventType.POSITION_CHANGED, t._onTargetPositionChange, t);
            t.target.on(cc.Node.EventType.SIZE_CHANGED, t._onTargetSizeChange, t);
            t.target.on(cc.Node.EventType.ANCHOR_CHANGED, t._onTargetAhchorChange, t);
        }
    },

    _unregisterEvent() {
        let t = this;
        cc.director.off(cc.Director.EVENT_AFTER_UPDATE, t._onEvAfterUpdate, t);
        t.node.off(cc.Node.EventType.POSITION_CHANGED, t._initNodeInfo, t);
        t.node.off(cc.Node.EventType.SIZE_CHANGED, t._initNodeInfo, t);
        if (t.target) {
            t.target.off(cc.Node.EventType.POSITION_CHANGED, t._onTargetPositionChange, t);
            t.target.off(cc.Node.EventType.SIZE_CHANGED, t._onTargetSizeChange, t);
            t.target.off(cc.Node.EventType.ANCHOR_CHANGED, t._onTargetAhchorChange, t);
        }
    },

    _initNodeInfo() {
        this._myLastPos = this.node.getPosition();
        this._myLastSize = this.node.getContentSize();
    },

    _initTargetInfo() {
        let t = this;
        if (t.target) {
            t._targetLastPos = t.target.getPosition();
            t._targetLastSize = t.target.getContentSize();
            t._targetLastAnchor = t.target.getAnchorPoint();
        }
    },

    _onTargetPositionChange(pos, force) {
        if (!this.enabled || !this.target || !pos || (!force && (pos.x == this._targetLastPos.x && pos.y == this._targetLastPos.y)))
            return;

        if (this.haveRelation(Type.LEFT_LEFT)
            || this.haveRelation(Type.LEFT_CENTER)
            || this.haveRelation(Type.LEFT_RIGHT)
            || this.haveRelation(Type.RIGHT_RIGHT)
            || this.haveRelation(Type.RIGHT_CENTER)
            || this.haveRelation(Type.RIGHT_LEFT)
            || this.haveRelation(Type.HORIZONTAL_CENTER)
        ) {
            this.node.x = this._myLastPos.x + (pos.x - this._targetLastPos.x);
        }

        if (this.haveRelation(Type.TOP_TOP)
            || this.haveRelation(Type.TOP_CENTER)
            || this.haveRelation(Type.TOP_BOTTOM)
            || this.haveRelation(Type.BOTTOM_BOTTOM)
            || this.haveRelation(Type.BOTTOM_CENTER)
            || this.haveRelation(Type.BOTTOM_TOP)
            || this.haveRelation(Type.VERTICAL_CENTER)
        ) {
            this.node.y = this._myLastPos.y + (pos.y - this._targetLastPos.y);
        }

        this._targetLastPos = pos;
    },

    _onTargetSizeChange(size, force) {
        if (!this.enabled || !this.target || !size || (!force && (size.width == this._targetLastSize.width && size.height == this._targetLastSize.height)))
            return;
        if (this.haveRelation(Type.WIDTH_WIDTH)) {
            this.node.width = this._myLastSize.width + (size.width - this._targetLastSize.width);
        }
        if (this.haveRelation(Type.HEIGHT_HEIGHT)) {
            this.node.height = this._myLastSize.height + (size.height - this._targetLastSize.height);
        }

        let targetLastL,
            targetNowL,
            targetLastR,
            targetNowR,
            targetLastB,
            targetNowB,
            targetLastT,
            targetNowT,
            targetLastC,
            targetNowC;

        if (this.haveRelation(Type.LEFT_LEFT)
            || this.haveRelation(Type.RIGHT_LEFT)
        ) {
            targetLastL = this._targetLastPos.x - (this._targetLastAnchor.x * this._targetLastSize.width);
            targetNowL = this.target.x - (this.target.anchorX * this.target.width);
            this.node.x = this._myLastPos.x + (targetNowL - targetLastL);
        }
        if (this.haveRelation(Type.RIGHT_RIGHT)
            || this.haveRelation(Type.LEFT_RIGHT)
        ) {
            targetLastR = this._targetLastPos.x + ((1 - this._targetLastAnchor.x) * this._targetLastSize.width);
            targetNowR = this.target.x + ((1 - this.target.anchorX) * this.target.width);
            this.node.x = this._myLastPos.x + (targetNowR - targetLastR);
        }
        if (this.haveRelation(Type.BOTTOM_BOTTOM)
            || this.haveRelation(Type.TOP_BOTTOM)
        ) {
            targetLastB = this._targetLastPos.y - (this._targetLastAnchor.y * this._targetLastSize.height);
            targetNowB = this.target.y - (this.target.anchorY * this.target.height);
            this.node.y = this._myLastPos.y + (targetNowB - targetLastB);
        }
        if (this.haveRelation(Type.TOP_TOP)
            || this.haveRelation(Type.BOTTOM_TOP)
        ) {
            targetLastT = this._targetLastPos.y + ((1 - this._targetLastAnchor.y) * this._targetLastSize.height);
            targetNowT = this.target.y + ((1 - this.target.anchorY) * this.target.height);
            this.node.y = this._myLastPos.y + (targetNowT - targetLastT);
        }

        if (this.haveRelation(Type.LEFT_CENTER)
            || this.haveRelation(Type.RIGHT_CENTER)
            || this.haveRelation(Type.HORIZONTAL_CENTER)
        ) {
            targetLastL = this._targetLastPos.x - (this._targetLastAnchor.x * this._targetLastSize.width);
            targetLastR = this._targetLastPos.x + ((1 - this._targetLastAnchor.x) * this._targetLastSize.width);
            targetLastC = (targetLastL + targetLastR) / 2;
            targetNowL = this.target.x - (this.target.anchorX * this.target.width);
            targetNowR = this.target.x + ((1 - this.target.anchorX) * this.target.width);
            targetNowC = (targetNowL + targetNowR) / 2;
            this.node.x = this._myLastPos.x + (targetNowC - targetLastC);
        }

        if (this.haveRelation(Type.TOP_CENTER)
            || this.haveRelation(Type.BOTTOM_CENTER)
            || this.haveRelation(Type.VERTICAL_CENTER)
        ) {
            targetLastB = this._targetLastPos.y - (this._targetLastAnchor.y * this._targetLastSize.height);
            targetLastT = this._targetLastPos.y + ((1 - this._targetLastAnchor.y) * this._targetLastSize.height);
            targetLastC = (targetLastB + targetLastT) / 2;
            targetNowB = this.target.y - (this.target.anchorY * this.target.height);
            targetNowT = this.target.y + ((1 - this.target.anchorY) * this.target.height);
            targetNowC = (targetNowB + targetNowT) / 2;
            this.node.y = this._myLastPos.y + (targetNowC - targetLastC);
        }

        this._targetLastSize = size;
    },

    _onTargetAhchorChange() {
        if (this.target) {
            this._onTargetSizeChange(this.target.getContentSize(), 1);
            this._targetLastAnchor = this.target.getAnchorPoint();
        }
    },

    _onEvAfterUpdate() {
        if (this.target) {
            this._onTargetPositionChange(this.target.getPosition());
            this._onTargetSizeChange(this.target.getContentSize());
        }
    },
    /**
     * 判断是否有关联
     * @param {Number} type
     * @returns
     */
    haveRelation(type) {
        return this.relations.indexOf(type) >= 0;
    },
    /**
    * 添加关联
    * @param {Number} type
    */
    addRelation(type) {
        if (!this.haveRelation(type))
            this.relations.push(type);
    },
    /**
    * 删除关联
    * @param {Number} type
    */
    delRelation(type) {
        let hasId = this.relations.indexOf(type) >= 0;
        if (hasId >= 0)
            this.relations.splice(hasId, 1);
    },
});